@@ -1,5 +1,5 @@
 // check-if-email-exists
-// Copyright (C) 2018-2021 Reacher
+// Copyright (C) 2018-2022 Reacher
 
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Affero General Public License as published
@@ -19,12 +19,13 @@ mod yahoo;
 use super::util::{constants::LOG_TARGET, input_output::CheckEmailInput};
 use crate::util::ser_with_display::ser_with_display;
 use async_native_tls::TlsConnector;
+use async_recursion::async_recursion;
 use async_smtp::{
 	smtp::{
 		client::net::NetworkStream, commands::*, error::Error as AsyncSmtpError,
 		extension::ClientId,
 	},
-	ClientSecurity, ClientTlsParameters, EmailAddress, SmtpClient, SmtpTransport,
+	ClientTlsParameters, EmailAddress, SmtpClient, SmtpTransport,
 };
 use async_std::future;
 use fast_socks5::{
@@ -34,6 +35,7 @@ use fast_socks5::{
 use rand::rngs::SmallRng;
 use rand::{distributions::Alphanumeric, Rng, SeedableRng};
 use serde::{Deserialize, Serialize};
+use std::default::Default;
 use std::iter;
 use std::str::FromStr;
 use std::time::Duration;
@@ -41,7 +43,7 @@ use trust_dns_proto::rr::Name;
 use yahoo::YahooError;
 
 /// Details that we gathered from connecting to this email via SMTP
-#[derive(Debug, Deserialize, Serialize)]
+#[derive(Debug, Default, Deserialize, Serialize)]
 pub struct SmtpDetails {
 	/// Are we able to connect to the SMTP server?
 	pub can_connect_smtp: bool,
@@ -55,18 +57,6 @@ pub struct SmtpDetails {
 	pub is_disabled: bool,
 }
 
-impl Default for SmtpDetails {
-	fn default() -> Self {
-		SmtpDetails {
-			can_connect_smtp: false,
-			has_full_inbox: false,
-			is_catch_all: false,
-			is_deliverable: false,
-			is_disabled: false,
-		}
-	}
-}
-
 /// Error occured connecting to this email server via SMTP.
 #[derive(Debug, Serialize)]
 #[serde(tag = "type", content = "message")]
@@ -74,8 +64,30 @@ pub enum SmtpError {
 	/// Error if we're using a SOCKS5 proxy.
 	#[serde(serialize_with = "ser_with_display")]
 	SocksError(SocksError),
+	/// Error when communicating with SMTP server, HELO phase.
+	#[serde(serialize_with = "ser_with_display")]
+	HeloError(AsyncSmtpError),
+	/// Error when communicating with SMTP server, connect phase.
+	#[serde(serialize_with = "ser_with_display")]
+	ConnectError(AsyncSmtpError),
+	/// Error when communicating with SMTP server, connect_with_stream phase.
+	#[serde(serialize_with = "ser_with_display")]
+	ConnectWithStreamError(AsyncSmtpError),
+	/// Error when communicating with SMTP server, MAIL FROM phase.
+	#[serde(serialize_with = "ser_with_display")]
+	MailFromError(AsyncSmtpError),
+	/// Error when communicating with SMTP server, RCPT TO phase.
+	#[serde(serialize_with = "ser_with_display")]
+	RcptToError(AsyncSmtpError),
+	/// Error when communicating with SMTP server, close phase.
+	#[serde(serialize_with = "ser_with_display")]
+	CloseError(AsyncSmtpError),
 	/// Error when communicating with SMTP server.
 	#[serde(serialize_with = "ser_with_display")]
+	#[deprecated(
+		since = "0.8.27",
+		note = "SMTP errors have been broken up into 6 phases, see all `Smtp*Error`s"
+	)]
 	SmtpError(AsyncSmtpError),
 	/// Time-out error.
 	#[serde(serialize_with = "ser_with_display")]
@@ -84,12 +96,6 @@ pub enum SmtpError {
 	YahooError(YahooError),
 }
 
-impl From<AsyncSmtpError> for SmtpError {
-	fn from(error: AsyncSmtpError) -> Self {
-		SmtpError::SmtpError(error)
-	}
-}
-
 impl From<SocksError> for SmtpError {
 	fn from(error: SocksError) -> Self {
 		SmtpError::SocksError(error)
@@ -110,12 +116,13 @@ impl From<YahooError> for SmtpError {
 
 /// Try to send an smtp command, close and return Err if fails.
 macro_rules! try_smtp (
-    ($res: expr, $client: ident, $host: expr, $port: expr) => ({
+    ($res: expr, $client: ident, $to_email: expr, $host: expr, $port: expr, $err_type: expr) => ({
 		if let Err(err) = $res {
-			log::debug!(target: LOG_TARGET, "Closing {}:{}, because of error '{}'.", $host, $port, err);
-			$client.close().await?;
+			log::debug!(target: LOG_TARGET, "email={} Closing {}:{}, because of error '{:?}'.", $to_email, $host, $port, err);
+			// Try to close the connection, but ignore if there's an error.
+			let _ = $client.close().await;
 
-			return Err(err.into());
+			return Err($err_type(err));
 		}
     })
 );
@@ -128,21 +135,28 @@ async fn connect_to_host(
 ) -> Result<SmtpTransport, SmtpError> {
 	let security = {
 		let host_string = host.to_utf8();
-		let host_str = host_string.trim_end_matches('.'); // hostname verification fails if it ends with '.'
-		ClientSecurity::Opportunistic(ClientTlsParameters::new(
-			host_str.to_string(),
-			TlsConnector::new().use_sni(true),
-		))
+		let host_string = host_string.trim_end_matches('.'); // hostname verification fails if it ends with '.'
+		let tls_params =
+			ClientTlsParameters::new(host_string.to_string(), TlsConnector::new().use_sni(true));
+
+		input.smtp_security.to_client_security(tls_params)
 	};
 	let mut smtp_client = SmtpClient::with_security((host.to_utf8().as_ref(), port), security)
-		.await?
+		.await
+		.map_err(SmtpError::HeloError)?
 		// FIXME Do not clone?
 		.hello_name(ClientId::Domain(input.hello_name.clone()))
 		.timeout(Some(Duration::new(30, 0))) // Set timeout to 30s
 		.into_transport();
 
 	// Connect to the host. If the proxy argument is set, use it.
-	log::debug!(target: LOG_TARGET, "Connecting to {}:{}", host, port);
+	log::debug!(
+		target: LOG_TARGET,
+		"email={} Connecting to {}:{}",
+		input.to_emails[0],
+		host,
+		port
+	);
 	if let Some(proxy) = &input.proxy {
 		let stream = Socks5Stream::connect(
 			(proxy.host.as_ref(), proxy.port),
@@ -157,11 +171,20 @@ async fn connect_to_host(
 				.connect_with_stream(NetworkStream::Socks5Stream(stream))
 				.await,
 			smtp_client,
+			input.to_emails[0],
 			host,
-			port
+			port,
+			SmtpError::ConnectWithStreamError
 		);
 	} else {
-		try_smtp!(smtp_client.connect().await, smtp_client, host, port);
+		try_smtp!(
+			smtp_client.connect().await,
+			smtp_client,
+			input.to_emails[0],
+			host,
+			port,
+			SmtpError::ConnectError
+		);
 	}
 
 	// "MAIL FROM: user@example.org"
@@ -178,8 +201,10 @@ async fn connect_to_host(
 			.command(MailCommand::new(Some(from_email.clone()), vec![],))
 			.await,
 		smtp_client,
+		input.to_emails[0],
 		host,
-		port
+		port,
+		SmtpError::MailFromError
 	);
 
 	Ok(smtp_client)
@@ -330,7 +355,7 @@ async fn email_deliverable(
 				});
 			}
 
-			Err(SmtpError::SmtpError(err))
+			Err(SmtpError::RcptToError(err))
 		}
 	}
 }
@@ -387,6 +412,11 @@ async fn create_smtp_future(
 		// so we can only check for "io: incomplete" SMTP error being returned.
 		// https://github.com/async-email/async-smtp/issues/37
 		if is_io_incomplete_smtp_error(&result) {
+			log::debug!(
+				target: LOG_TARGET,
+				"Got `io: incomplete` error, reconnecting."
+			);
+
 			let _ = smtp_client.close().await;
 			smtp_client = connect_to_host(host, port, input).await?;
 			result = email_deliverable(&mut smtp_client, to_email).await;
@@ -395,23 +425,28 @@ async fn create_smtp_future(
 		result?
 	};
 
-	smtp_client.close().await?;
+	smtp_client.close().await.map_err(SmtpError::CloseError)?;
 
 	Ok((is_catch_all, deliverability))
 }
 
 /// Indicates whether the given [`Result`] represents an `io: incomplete`
-/// [`SmtpError::SmtpError`].
+/// [`SmtpError::Error`].
 fn is_io_incomplete_smtp_error<T>(result: &Result<T, SmtpError>) -> bool {
-	if let Err(SmtpError::SmtpError(AsyncSmtpError::Io(err))) = result {
-		err.to_string().as_str() == "incomplete"
-	} else {
-		false
+	match result {
+		Err(SmtpError::HeloError(AsyncSmtpError::Io(err)))
+		| Err(SmtpError::ConnectError(AsyncSmtpError::Io(err)))
+		| Err(SmtpError::ConnectWithStreamError(AsyncSmtpError::Io(err)))
+		| Err(SmtpError::MailFromError(AsyncSmtpError::Io(err)))
+		| Err(SmtpError::RcptToError(AsyncSmtpError::Io(err)))
+		| Err(SmtpError::CloseError(AsyncSmtpError::Io(err))) => err.to_string().as_str() == "incomplete",
+		_ => false,
 	}
 }
 
-/// Get all email details we can from one single `EmailAddress`.
-pub async fn check_smtp(
+/// Get all email details we can from one single `EmailAddress`, without
+/// retries.
+async fn check_smtp_without_retry(
 	to_email: &EmailAddress,
 	host: &Name,
 	port: u16,
@@ -441,6 +476,60 @@ pub async fn check_smtp(
 	})
 }
 
+/// Get all email details we can from one single `EmailAddress`.
+/// Retry the SMTP connection, in particular to avoid greylisting.
+#[async_recursion]
+async fn retry(
+	to_email: &EmailAddress,
+	host: &Name,
+	port: u16,
+	domain: &str,
+	input: &CheckEmailInput,
+	count: usize,
+) -> Result<SmtpDetails, SmtpError> {
+	log::debug!(
+		target: LOG_TARGET,
+		"email={} Check SMTP attempt {}/{}",
+		input.to_emails[0],
+		input.retries - count + 1,
+		input.retries,
+	);
+
+	let result = check_smtp_without_retry(to_email, host, port, domain, input).await;
+
+	log::debug!(
+		target: LOG_TARGET,
+		"email={} Got result with attempt {}/{}, result={:?}",
+		input.to_emails[0],
+		input.retries - count + 1,
+		input.retries,
+		result
+	);
+
+	match result {
+		Ok(_) => result,
+		Err(_) => {
+			if count <= 1 {
+				result
+			} else {
+				retry(to_email, host, port, domain, input, count - 1).await
+			}
+		}
+	}
+}
+
+/// Get all email details we can from one single `EmailAddress`, without
+/// retries.
+pub async fn check_smtp(
+	to_email: &EmailAddress,
+	host: &Name,
+	port: u16,
+	domain: &str,
+	input: &CheckEmailInput,
+) -> Result<SmtpDetails, SmtpError> {
+	retry(to_email, host, port, domain, input, input.retries).await
+}
+
 #[cfg(test)]
 mod tests {
 	use super::{check_smtp, CheckEmailInput, SmtpError};
