@@ -1,5 +1,5 @@
 // check-if-email-exists
-// Copyright (C) 2018-2021 Reacher
+// Copyright (C) 2018-2022 Reacher
 
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Affero General Public License as published
@@ -18,6 +18,7 @@ use crate::misc::{MiscDetails, MiscError};
 use crate::mx::{MxDetails, MxError};
 use crate::smtp::{SmtpDetails, SmtpError};
 use crate::syntax::SyntaxDetails;
+use async_smtp::{ClientSecurity, ClientTlsParameters};
 use serde::{ser::SerializeMap, Deserialize, Serialize, Serializer};
 use std::time::Duration;
 
@@ -31,6 +32,31 @@ pub struct CheckEmailInputProxy {
 	pub port: u16,
 }
 
+/// Define how to apply TLS to a SMTP client connection. Will be converted into
+/// async_smtp::ClientSecurity.
+#[derive(Debug, Clone, Copy, Deserialize, Serialize)]
+pub enum SmtpSecurity {
+	/// Insecure connection only (for testing purposes).
+	None,
+	/// Start with insecure connection and use `STARTTLS` when available.
+	Opportunistic,
+	/// Start with insecure connection and require `STARTTLS`.
+	Required,
+	/// Use TLS wrapped connection.
+	Wrapper,
+}
+
+impl SmtpSecurity {
+	pub fn to_client_security(self, tls_params: ClientTlsParameters) -> ClientSecurity {
+		match self {
+			Self::None => ClientSecurity::None,
+			Self::Opportunistic => ClientSecurity::Opportunistic(tls_params),
+			Self::Required => ClientSecurity::Required(tls_params),
+			Self::Wrapper => ClientSecurity::Wrapper(tls_params),
+		}
+	}
+}
+
 /// Builder pattern for the input argument into the main `email_exists`
 /// function.
 #[derive(Debug, Clone, Deserialize, Serialize)]
@@ -60,6 +86,14 @@ pub struct CheckEmailInput {
 	///
 	/// Defaults to true.
 	pub yahoo_use_api: bool,
+	/// Number of retries of SMTP connections to do.
+	///
+	/// Defaults to 2 to avoid greylisting.
+	pub retries: usize,
+	/// How to apply TLS to a SMTP client connection.
+	///
+	/// Defaults to Opportunistic.
+	pub smtp_security: SmtpSecurity,
 }
 
 impl Default for CheckEmailInput {
@@ -70,8 +104,10 @@ impl Default for CheckEmailInput {
 			hello_name: "localhost".into(),
 			proxy: None,
 			smtp_port: 25,
+			smtp_security: SmtpSecurity::Opportunistic,
 			smtp_timeout: None,
 			yahoo_use_api: true,
+			retries: 2,
 		}
 	}
 }
@@ -131,6 +167,12 @@ impl CheckEmailInput {
 		self
 	}
 
+	/// Set the number of SMTP retries to do.
+	pub fn set_retries(&mut self, retries: usize) -> &mut CheckEmailInput {
+		self.retries = retries;
+		self
+	}
+
 	/// Add optional timeout for the SMTP verification step.
 	#[deprecated(since = "0.8.24", note = "Please use set_smtp_timeout instead")]
 	pub fn smtp_timeout(&mut self, duration: Duration) -> &mut CheckEmailInput {
@@ -144,6 +186,12 @@ impl CheckEmailInput {
 		self
 	}
 
+	/// Set the SMTP client security to use for TLS.
+	pub fn set_smtp_security(&mut self, smtp_security: SmtpSecurity) -> &mut CheckEmailInput {
+		self.smtp_security = smtp_security;
+		self
+	}
+
 	/// Add optional timeout for the SMTP verification step.
 	pub fn set_smtp_timeout(&mut self, duration: Duration) -> &mut CheckEmailInput {
 		self.smtp_timeout = Some(duration);
